import numpy as np
import random
import math

from une_ai.models.grid_map import GridMap
from une_ai.models.environment import Environment

class VacuumEnvironment(Environment):

    def __init__(self, w, h):
        super().__init__('Vacuum world')
        self._w = w
        self._h = h
        self._dirt_map = GridMap(w, h)
        self._walls_map = GridMap(w, h)
        self._explored_tiles = GridMap(w, h)
        self._initialise_walls(0.01)
        self._initialise_dirt(0.01)

        self._score = 0
    
    def get_percepts(self, agent):
        # starts the percepts with current values of sensors
        percepts = agent.read_sensors()
        agent_x, agent_y = percepts['location-sensor']
        coords = {
            'center': (agent_x, agent_y),
            'north': (agent_x, agent_y-1),
            'south': (agent_x, agent_y+1),
            'west': (agent_x-1, agent_y),
            'east': (agent_x+1, agent_y)
        }

        # reset bumper sensors
        for direction in coords.keys():
            if direction != 'center':
                percepts['bumper-sensor-{0}'.format(direction)] = False
        
        # check if the agent is crashing against a wall
        cur_direction = agent.read_actuator_value('wheels-direction')
        if self.is_wall(agent_x, agent_y):
            # there is a wall!
            # set agent location sensor to past x, y
            if cur_direction == 'north':
                past_x, past_y = coords['south']
            elif cur_direction == 'south':
                past_x, past_y = coords['north']
            elif cur_direction == 'west':
                past_x, past_y = coords['east']
            elif cur_direction == 'east':
                past_x, past_y = coords['west']
            
            
            percepts['location-sensor'] = (past_x, past_y)

            # update the bumper sensor
            percepts['bumper-sensor-{0}'.format(cur_direction)] = True
        
        # now check if there is dirt in the surroundings
        for coord_name, coord in coords.items():
            try:
                is_dirty = self.is_dirty(coord[0], coord[1])
            except:
                is_dirty = False
            percepts['dirt-sensor-{0}'.format(coord_name)] = is_dirty
            
        return percepts
    
    def get_score(self):
        return self._score

    def get_width(self):
        return self._w
    
    def get_height(self):
        return self._h
    
    def set_explored(self, pos_x, pos_y):
        self._explored_tiles.set_item_value(pos_x, pos_y, True)
        
    def _initialise_walls(self, wall_density=0.01):
        total_size = self._w*self._h
        n_walls = math.floor(total_size*wall_density)
        empty_tiles = self._walls_map.find_value(False)
        walls_coords = random.sample(empty_tiles, n_walls)
        for coord in walls_coords:
            orientation = random.choice(['h', 'v'])
            length = random.choice([3, 4, 5])
            cur_x = coord[0]
            cur_y = coord[1]
            for i in range(0, length):
                if cur_x >= self._w or cur_y >= self._h:
                    #if out of bounds, end
                    break
                self._walls_map.set_item_value(cur_x, cur_y, True)
                if orientation == 'h':
                    cur_x += 1
                else:
                    cur_y += 1
    
    def _initialise_dirt(self, dirt_density=0.01):
        total_size = self._w*self._h
        n_dirt = math.floor(total_size*dirt_density)
        self._spawn_dirt(n_dirt)
            
    def _spawn_dirt(self, k=1):
        condition_func = lambda walls_map: np.logical_and(walls_map == False, self._dirt_map.get_map() == False)
        valid_coords = self._walls_map.find_value_by_condition(condition_func)
        new_coord = random.sample(valid_coords, k)
        for coord in new_coord:
            self._dirt_map.set_item_value(coord[0], coord[1], True)

    def is_dirty(self, pos_x, pos_y):
        return self._dirt_map.get_item_value(pos_x, pos_y)
    
    def is_wall(self, pos_x, pos_y):
        try:
            return self._walls_map.get_item_value(pos_x, pos_y)
        except:
            # out of bounds, it's a wall
            return True
    
    def remove_dirt(self, x, y):
        try:
            is_dirt = self.is_dirty(x, y)
            if is_dirt:
                self._score += 1
                self._dirt_map.set_item_value(x, y, False)
        except:
            # we may be out of bounds, but there is no dirt outside bounds
            pass
    
    def get_dirt_coords(self):
        return self._dirt_map.find_value(True)
    
    def get_walls_coords(self):
        return self._walls_map.find_value(True)
    
    def get_free_coords(self):
        condition_func = lambda walls_map: np.logical_and(walls_map == False, self._dirt_map.get_map() == False)
        return self._walls_map.find_value_by_condition(condition_func)
    
    def get_explored_tiles_coords(self):
        return self._explored_tiles.find_value(True)


import random
import numpy as np

from une_ai.models import Agent

class VacuumAgent(Agent):

    WHEELS_DIRECTIONS = ['north', 'south', 'west', 'east']

    def __init__(self, agent_program, init_x=0, init_y=0, max_battery_level=1000):
        super().__init__(
            agent_name='vacuum_agent',
            agent_program=agent_program
        )

        self._collision = False
        self._max_battery_level = max_battery_level

        # adding the sensors
        self.add_sensor('battery-level', max_battery_level, lambda v: isinstance(v, int) and v >= 0)
        self.add_sensor('location-sensor', (init_x, init_y), lambda v: isinstance(v, tuple) and isinstance(v[0], int) and isinstance(v[1], int))

        directions = VacuumAgent.WHEELS_DIRECTIONS.copy()
        directions.append('center')
        for direction in directions:
            self.add_sensor('dirt-sensor-{0}'.format(direction), False, lambda v: isinstance(v, bool) or isinstance(v, np.bool_))
            if direction != 'center':
                self.add_sensor('bumper-sensor-{0}'.format(direction), False, lambda v: isinstance(v, bool) or isinstance(v, np.bool_))

        # adding the actuators
        self.add_actuator(
            'wheels-direction',
            random.choice(VacuumAgent.WHEELS_DIRECTIONS),
            lambda v: v in VacuumAgent.WHEELS_DIRECTIONS
        )
        self.add_actuator(
            'vacuum-power',
            0,
            lambda v: v in [0, 1]
        )
        self.add_actuator(
            'suction-power',
            0,
            lambda v: v in [0, 1]
        )

        # adding the actions
        self.add_action(
            'start-cleaning',
            lambda: {'vacuum-power': 1} if not self.is_out_of_charge() else {}
        )
        self.add_action(
            'stop-cleaning',
            lambda: {
                'vacuum-power': 0
            }
        )
        self.add_action(
            'activate-suction-mechanism',
            lambda: {'suction-power': 1} if not self.is_out_of_charge() else {}
        )
        self.add_action(
            'deactivate-suction-mechanism',
            lambda: {
                'suction-power': 0
            }
        )
        for direction in VacuumAgent.WHEELS_DIRECTIONS:
            self.add_action(
                'change-direction-{0}'.format(direction),
                lambda d=direction: {'wheels-direction': d} if not self.is_out_of_charge() else {}
            )

    def get_pos_x(self):
        return self.read_sensor_value('location-sensor')[0]
    
    def get_pos_y(self):
        return self.read_sensor_value('location-sensor')[1]

    def _update_position(self, new_x, new_y):
        self._pos_x = new_x
        self._pos_y = new_y
        self._update_sensor_value('location-sensor', (new_x, new_y))
    
    def _consume_battery(self, consumption=1):
        cur_level = self.read_sensor_value('battery-level')
        if cur_level - consumption < 0:
            new_level = 0

            # turn off powers
            self._update_actuator_value('suction-power', 0)
            self._update_actuator_value('vacuum-power', 0)
        else:
            new_level = cur_level - consumption
        
        self._update_sensor_value('battery-level', new_level)
    
    def is_out_of_charge(self):
        return self.read_sensor_value('battery-level') == 0
    
    def get_battery_level(self):
        return int(self.read_sensor_value('battery-level')/self._max_battery_level*100)
    
    def collision_detected(self):
        self._collision = True
    
    def reset_collisions(self):
        self._collision = False

    def did_collide(self):
        return self._collision

    def _interact_with_environment(self, environment):
        self.reset_collisions()

        # remove dirt on current position if suction power on
        cur_x = self.get_pos_x()
        cur_y = self.get_pos_y()
        
        if self.read_actuator_value('suction-power') == 1:
            environment.remove_dirt(cur_x, cur_y)
            self._consume_battery(5)

        if self.read_actuator_value('vacuum-power') == 1:
            self._consume_battery(1)
            direction = self.read_actuator_value('wheels-direction')
            if direction == "west":
                offset = (-1, 0)
            elif direction == "east":
                offset = (1, 0)
            elif direction == "north":
                offset = (0, -1)
            elif direction == "south":
                offset = (0, 1)
            
            new_x, new_y = (cur_x + offset[0], cur_y + offset[1])
            
            try:
                environment.set_explored(new_x, new_y)
            except:
                # out of boundaries, but that's all good
                # the environment class will manage the
                # crash with the wall when sensing
                pass

            if environment.is_wall(new_x, new_y):
                self.collision_detected()

            self._update_position(new_x, new_y)
                